# coding=utf-8
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
@file knn.py_in

@brief knn: K-Nearest Neighbors for regression and classification

@namespace knn

"""

import plpy
from utilities.validate_args import input_tbl_valid, output_tbl_valid
from utilities.validate_args import cols_in_tbl_valid
from utilities.validate_args import is_col_array
from utilities.validate_args import array_col_has_no_null
from utilities.validate_args import get_expr_type
from utilities.utilities import _assert
from utilities.utilities import unique_string
from utilities.control import MinWarning
from utilities.validate_args import quote_ident
from utilities.validate_args import is_var_valid
from utilities.utilities import NUMERIC, ONLY_ARRAY
from utilities.utilities import is_valid_psql_type

MAX_WEIGHT_ZERO_DIST = 1e6


def knn_validate_src(schema_madlib, point_source, point_column_name, point_id,
                     label_column_name, test_source, test_column_name,
                     test_id, output_table, k, output_neighbors, fn_dist,
                     **kwargs):
    input_tbl_valid(point_source, 'kNN')
    input_tbl_valid(test_source, 'kNN')
    output_tbl_valid(output_table, 'kNN')

    _assert(label_column_name or output_neighbors,
            "kNN error: Either label_column_name or "
            "output_neighbors has to be inputed.")

    if label_column_name and label_column_name.strip():
        cols_in_tbl_valid(point_source, [label_column_name], 'kNN')

    _assert(is_var_valid(point_source, point_column_name),
            "kNN error: {0} is an invalid column name or expression for point_column_name param".format(point_column_name))
    point_col_type = get_expr_type(point_column_name, point_source)
    _assert(is_valid_psql_type(point_col_type, NUMERIC | ONLY_ARRAY),
            "kNN Error: Feature column or expression '{0}' in train table is not"
            " an array.".format(point_column_name))

    _assert(is_var_valid(test_source, test_column_name),
            "kNN error: {0} is an invalid column name or expression for test_column_name param".format(test_column_name))
    test_col_type = get_expr_type(test_column_name, test_source)
    _assert(is_valid_psql_type(test_col_type, NUMERIC | ONLY_ARRAY),
            "kNN Error: Feature column or expression '{0}' in test table is not"
            " an array.".format(test_column_name))

    cols_in_tbl_valid(point_source, [point_id], 'kNN')
    cols_in_tbl_valid(test_source, [test_id], 'kNN')

    if not array_col_has_no_null(point_source, point_column_name):
        plpy.error("kNN Error: Feature column '{0}' in train table has some"
                   " NULL values.".format(point_column_name))
    if not array_col_has_no_null(test_source, test_column_name):
        plpy.error("kNN Error: Feature column '{0}' in test table has some"
                   " NULL values.".format(test_column_name))

    if k <= 0:
        plpy.error("kNN Error: k={0} is an invalid value, must be greater "
                   "than 0.".format(k))

    bound = plpy.execute("SELECT {k} <= count(*) AS bound FROM {tbl}".
                         format(k=k, tbl=point_source))[0]['bound']
    if not bound:
        plpy.error("kNN Error: k={0} is greater than number of rows in"
                   " training table.".format(k))

    if label_column_name:
        col_type = get_expr_type(label_column_name, point_source).lower()
        if col_type not in ['integer', 'double precision', 'float', 'boolean']:
            plpy.error("kNN error: Invalid data type '{0}' for"
                       " label_column_name in table '{1}'.".
                       format(col_type, point_source))

    col_type_test = get_expr_type(test_id, test_source).lower()
    if col_type_test not in ['integer']:
        plpy.error("kNN Error: Invalid data type '{0}' for"
                   " test_id column in table '{1}'.".
                   format(col_type_test, test_source))

    if fn_dist:
        fn_dist = fn_dist.lower().strip()
        dist_functions = set(["{0}.{1}".format(schema_madlib, dist) for dist in
                              ('dist_norm1', 'dist_norm2',
                               'squared_dist_norm2', 'dist_angle',
                               'dist_tanimoto')])

        is_invalid_func = plpy.execute("""
            SELECT prorettype != 'DOUBLE PRECISION'::regtype OR
                   proisagg = TRUE AS OUTPUT
            FROM pg_proc
            WHERE oid='{fn_dist}(DOUBLE PRECISION[], DOUBLE PRECISION[])'::regprocedure;
            """.format(fn_dist=fn_dist))[0]['output']

        if is_invalid_func or (fn_dist not in dist_functions):
            plpy.error("KNN error: Distance function has invalid signature "
                       "or is not a simple function.")

    return k
# ------------------------------------------------------------------------------


def knn(schema_madlib, point_source, point_column_name, point_id,
        label_column_name, test_source, test_column_name, test_id, output_table,
        k, output_neighbors, fn_dist, weighted_avg, **kwargs):
    """
        KNN function to find the K Nearest neighbours
        Args:
            @param schema_madlib        Name of the Madlib Schema
            @param point_source         Training data table
            @param point_column_name    Name of the column with training data
                                        or expression that evaluates to a
                                        numeric array
            @param point_id             Name of the column having ids of data
                                        point in train data table
                                        points.
            @param label_column_name    Name of the column with labels/values
                                        of training data points.
            @param test_source          Name of the table containing the test
                                        data points.
            @param test_column_name     Name of the column with testing data
                                        points or expression that evaluates to a
                                        numeric array
            @param test_id              Name of the column having ids of data
                                        points in test data table.
            @param output_table         Name of the table to store final
                                        results.
            @param k                    default: 1. Number of nearest
                                        neighbors to consider
            @output_neighbours          Outputs the list of k-nearest neighbors
                                        that were used in the voting/averaging.
            @param fn_dist              Distance metrics function. Default is
                                        squared_dist_norm2. Following functions
                                        are supported :
                                        dist_norm1 , dist_norm2,squared_dist_norm2,
                                        dist_angle , dist_tanimoto
                                        Or user defined function with signature
                                        DOUBLE PRECISION[] x, DOUBLE PRECISION[] y -> DOUBLE PRECISION
            @param weighted_avg         Calculates the Regression or classication of k-NN using
                                        the weighted average method.
    """
    with MinWarning('warning'):
        output_neighbors = True if output_neighbors is None else output_neighbors
        if k is None:
            k = 1
        knn_validate_src(schema_madlib, point_source,
                         point_column_name, point_id, label_column_name,
                         test_source, test_column_name, test_id,
                         output_table, k, output_neighbors, fn_dist)

        # Unique Strings
        x_temp_table = unique_string(desp='x_temp_table')
        y_temp_table = unique_string(desp='y_temp_table')
        label_col_temp = unique_string(desp='label_col_temp')
        test_id_temp = unique_string(desp='test_id_temp')
        train = unique_string(desp='train')
        test = unique_string(desp='test')
        p_col_name = unique_string(desp='p_col_name')
        t_col_name = unique_string(desp='t_col_name')
        dist = unique_string(desp='dist')
        train_id = unique_string(desp='train_id')
        dist_inverse = unique_string(desp='dist_inverse')
        r = unique_string(desp='r')

        if not fn_dist:
            fn_dist = '{0}.squared_dist_norm2'.format(schema_madlib)

        fn_dist = fn_dist.lower().strip()
        interim_table = unique_string(desp='interim_table')

        pred_out = ""
        knn_neighbors = ""
        label_out = ""
        cast_to_int = ""
        view_def = ""
        view_join = ""
        view_grp_by = ""

        if label_column_name:
            label_column_type = get_expr_type(
                label_column_name, point_source).lower()
            if label_column_type in ['boolean', 'integer', 'text']:
                is_classification = True
                cast_to_int = '::INTEGER'
            else:
                is_classification = False

            if is_classification:
                if weighted_avg:
                    # This view is to calculate the max value of sum of the 1/distance grouped by label and Id.
                    # And this max value will be the prediction for the
                    # classification model.
                    view_def = """
                        WITH vw AS (
                            SELECT DISTINCT ON({test_id_temp})
                                {test_id_temp},
                                last_value(data_sum) OVER (
                                    PARTITION BY {test_id_temp}
                                    ORDER BY data_sum, {label_col_temp}
                                    ROWS BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING
                                    ) AS data_dist ,
                                last_value({label_col_temp}) OVER (
                                    PARTITION BY {test_id_temp}
                                    ORDER BY data_sum, {label_col_temp}
                                    ROWS BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING
                                    ) AS {label_col_temp}
                            FROM   (
                                SELECT
                                    {test_id_temp},
                                    {label_col_temp},
                                    sum({dist_inverse}) data_sum
                                FROM pg_temp.{interim_table}
                                GROUP BY {test_id_temp},
                                         {label_col_temp}
                            ) a
                            -- GROUP BY {test_id_temp} , {label_col_temp}
                        )
                        """.format(**locals())
                    # This join is needed to get the max value of predicion
                    # calculated above
                    view_join = (" JOIN vw ON knn_temp.{0} = vw.{0}".
                                 format(test_id_temp))
                    view_grp_by = ", vw.{0}".format(label_col_temp)
                    pred_out = ", vw.{0}".format(label_col_temp)
                else:
                    pred_out = ", {0}.mode({1})".format(
                        schema_madlib, label_col_temp)
            else:
                if weighted_avg:
                    pred_out = (", sum({0} * {dist_inverse}) / sum({dist_inverse})".
                                format(label_col_temp, dist_inverse=dist_inverse))
                else:
                    pred_out = ", avg({0})".format(label_col_temp)

            pred_out += " AS prediction"
            label_out = (", {train}.{label_column_name}{cast_to_int}"
                         " AS {label_col_temp}").format(**locals())
            comma_label_out_alias = ', ' + label_col_temp
            label_name = ", {label_column_name}".format(
                label_column_name=label_column_name)

        else:
            pred_out = ""
            label_out = ""
            comma_label_out_alias = ""
            label_name = ""

        # interim_table picks the 'k' nearest neighbors for each test point
        if output_neighbors:
            knn_neighbors = (", array_agg(knn_temp.{train_id} ORDER BY "
                             "knn_temp.{dist_inverse} DESC) AS k_nearest_neighbours ").format(**locals())
        else:
            knn_neighbors = ''
        plpy.execute("""
            CREATE TEMP TABLE {interim_table} AS
                SELECT * FROM (
                    SELECT row_number() over
                            (partition by {test_id_temp} order by {dist}) AS {r},
                            {test_id_temp},
                            {train_id},
                            CASE WHEN {dist} = 0.0 THEN {max_weight_zero_dist}
                                 ELSE 1.0 / {dist}
                            END AS {dist_inverse}
                            {comma_label_out_alias}
                    FROM (
                        SELECT {test}.{test_id} AS {test_id_temp},
                            {train}.{point_id} as {train_id},
                            {fn_dist}(
                                {p_col_name},
                                {t_col_name})
                            AS {dist}
                            {label_out}
                            FROM
                            (
                            SELECT {point_id} , {point_column_name} as {p_col_name} {label_name} from {point_source}
                            ) {train},
                            (
                            SELECT {test_id} ,{test_column_name} as {t_col_name} from {test_source}
                            ) {test}
                        ) {x_temp_table}
                    ) {y_temp_table}
            WHERE {y_temp_table}.{r} <= {k}
            """.format(max_weight_zero_dist=MAX_WEIGHT_ZERO_DIST, **locals()))

        sql = """
            CREATE TABLE {output_table} AS
                {view_def}
                SELECT
                    knn_temp.{test_id_temp} AS id,
                    {test_column_name} as "{test_column_name}"
                    {pred_out}
                    {knn_neighbors}
                FROM
                    pg_temp.{interim_table}  AS knn_temp
                    JOIN
                    {test_source} AS knn_test
                ON knn_temp.{test_id_temp} = knn_test.{test_id}
                    {view_join}
                GROUP BY knn_temp.{test_id_temp},
                    {test_column_name}
                         {view_grp_by}
            """
        plpy.execute(sql.format(**locals()))
        plpy.execute("DROP TABLE IF EXISTS {0}".format(interim_table))
        return


def knn_help(schema_madlib, message, **kwargs):
    """
    Help function for knn

    Args:
        @param schema_madlib
        @param message: string, Help message string
        @param kwargs

    Returns:
        String. Help/usage information
    """
    if message is not None and \
            message.lower() in ("usage", "help", "?"):
        help_string = """
-----------------------------------------------------------------------
                            USAGE
-----------------------------------------------------------------------
SELECT {schema_madlib}.knn(
    point_source,       -- Training data table having training features as vector column and labels
    point_column_name,  -- Name of column having feature vectors in training data table
    point_id,           -- Name of column having feature vector Ids in train data table
    label_column_name,  -- Name of column having actual label/vlaue for corresponding feature vector in training data table
    test_source,        -- Test data table having features as vector column. Id of features is mandatory
    test_column_name,   -- Name of column having feature vectors in test data table
    test_id,     -- Name of column having feature vector Ids in test data table
    output_table,       -- Name of output table
    k,                  -- value of k. Default will go as 1
    output_neighbors    -- Outputs the list of k-nearest neighbors that were used in the voting/averaging.
    fn_dist             -- The name of the function to use to calculate the distance from a data point to a centroid.
    weighted_avg         Calculates the Regression or classication of k-NN using the weighted average method.
    );

-----------------------------------------------------------------------
                            OUTPUT
-----------------------------------------------------------------------
The output of the KNN module is a table with the following columns:

id                  The ids of test data points.
test_column_name    The test data points.
prediction          The output of KNN- label in case of classification, average value in case of regression.
k_nearest_neighbours The list of k-nearest neighbors that were used in the voting/averaging.
"""
    else:
        help_string = """
----------------------------------------------------------------------------
                                SUMMARY
----------------------------------------------------------------------------
k-Nearest Neighbors is a method for finding k closest points to a given data
point in terms of a given metric. Its input consist of data points as features
from testing examples. For a given k, it looks for k closest points in
training set for each of the data points in test set. Algorithm generates one
output per testing example. The output of KNN depends on the type of task:
For Classification, the output is majority vote of the classes of the k
nearest data points. The testing example gets assigned the most popular class
among nearest neighbors. For Regression, the output is average of the values
of k nearest neighbors of the given testing example.
--
For an overview on usage, run:
SELECT {schema_madlib}.knn('usage');
"""

    return help_string.format(schema_madlib=schema_madlib)
# ------------------------------------------------------------------------------
